#include "htmlretrieve.hpp"
#include "commandline.hpp"
#include <iostream>
#include <string>
#include <pugixml.hpp>
#include <sstream>
#include <fstream>
#include <utility>
#include <ciso646>
#include <memory>
#include <iterator>
#include <stdexcept>

namespace {
	typedef std::pair<int, int> LineColType;

	LineColType line_col_from_offset ( ptrdiff_t parOffset, const std::string& parData );
	void dump_string ( const std::string& parPathDest, const std::string& parData );
} //unnamed namespace

int main (int argc, char* argv[]) {
	using boost::program_options::variables_map;

	variables_map vm;
	try {
		if (duck::parse_commandline(argc, argv, vm)) {
			return 0;
		}
	}
	catch (const std::invalid_argument& err) {
		std::cerr << err.what() << "\nUse --help for help" << std::endl;
		return 2;
	}

	const auto url = vm["input-url"].as<std::string>();
	const auto xpath = vm["xpath"].as<std::string>();
#if !defined(NDEBUG)
	std::cout << "URL  : " << url << "\n";
	std::cout << "XPath: " << xpath << std::endl;
	std::cout << "Agent: " << vm["agent"].as<std::string>() << std::endl;
#endif

	std::string html;

	if ("-" != url) {
		html = duck::fetch_html(url, vm["agent"].as<std::string>(), false, false);
	}
	else {
		std::cin >> std::noskipws;
		std::istream_iterator<char> it(std::cin);
		std::istream_iterator<char> end;
		html = std::string(it, end);
	}

	if (vm.count("dump-raw")) {
		dump_string(vm["dump-raw"].as<std::string>(), html);
	}

	html = duck::clean_html(std::move(html));
	if (vm.count("dump")) {
		dump_string(vm["dump"].as<std::string>(), html);
	}

	{
		pugi::xml_document doc;
		std::istringstream iss(html);
		pugi::xml_parse_result result(doc.load(iss));
		if (not result) {
			auto line_col = line_col_from_offset(result.offset, html);
			std::cerr << "Error parsing the source XML at line " <<
				line_col.first << " col " << line_col.second << ":\n" <<
				result.description() << std::endl;
			return 1;
		}

		pugi::xpath_node_set xpathRes = doc.select_nodes(xpath.c_str());
		for (pugi::xpath_node_set::const_iterator itFind(xpathRes.begin()), itFindEND(xpathRes.end()); itFind != itFindEND; ++itFind) {
			const pugi::xpath_node& node = *itFind;
			if (node.node()) {
				std::cout << node.node().name() << ": " << node.node().value() << "\n";
			}
			else if (node.attribute()) {
				std::cout << node.attribute().name() << ": " << node.attribute().value() << "\n";
			}
		}
	}
	return 0;
}

namespace {
	LineColType line_col_from_offset (ptrdiff_t parOffset, const std::string& parData) {
		size_t index = 0;
		int line = 1;
		int chara = 1;
		while (parOffset and index < parData.size()) {
			if (parData[index] == '\n') {
				chara = 1;
				++line;
			}
			else {
				++chara;
			}
			++index;
			--parOffset;
		}
		return std::make_pair(line, chara);
	}

	void dump_string (const std::string& parPathDest, const std::string& parData) {
		std::unique_ptr<std::ofstream> ofs;
		const bool use_stdout = ("-" == parPathDest);
		if (not use_stdout) {
			ofs.reset(new std::ofstream(parPathDest));
		}
		std::ostream* const os = (use_stdout ? &std::cout : ofs.get());
		*os << parData;
	}
} //unnamed namespace
